@@ -110,3 +110,19 @@ class ShouldMatchB008RuleOfImmutableTypeAnnotationIgnored:
     # ignored
     this_is_fine: int = f()
 
+
+# Test for:
+# https://github.com/astral-sh/ruff/issues/17424
+@dataclass(frozen=True)
+class C:
+    foo: int = 1
+
+
+@dataclass
+class D:
+    c: C = C()
+
+
+@dataclass
+class E:
+    c: C = C()
\ No newline at end of file
