@@ -5,8 +5,8 @@ use std::path::PathBuf;
 use annotate::annotate_imports;
 use block::{Block, Trailer};
 pub(crate) use categorize::categorize;
+use categorize::categorize_imports;
 pub use categorize::{ImportSection, ImportType};
-use categorize::{MatchSourceStrategy, categorize_imports};
 use comments::Comment;
 use normalize::normalize_imports;
 use order::order_imports;
@@ -76,7 +76,6 @@ pub(crate) fn format_imports(
     source_type: PySourceType,
     target_version: PythonVersion,
     settings: &Settings,
-    match_source_strategy: MatchSourceStrategy,
     tokens: &Tokens,
 ) -> String {
     let trailer = &block.trailer;
@@ -104,7 +103,6 @@ pub(crate) fn format_imports(
             package,
             target_version,
             settings,
-            match_source_strategy,
         );
 
         if !block_output.is_empty() && !output.is_empty() {
@@ -161,7 +159,6 @@ fn format_import_block(
     package: Option<PackageRoot<'_>>,
     target_version: PythonVersion,
     settings: &Settings,
-    match_source_strategy: MatchSourceStrategy,
 ) -> String {
     #[derive(Debug, Copy, Clone, PartialEq, Eq)]
     enum LineInsertion {
@@ -182,7 +179,6 @@ fn format_import_block(
         settings.no_sections,
         &settings.section_order,
         &settings.default_section,
-        match_source_strategy,
     );
 
     let mut output = String::new();
@@ -797,6 +793,7 @@ mod tests {
     #[test_case(Path::new("docstring_followed_by_continuation.py"))]
     #[test_case(Path::new("docstring_only.py"))]
     #[test_case(Path::new("docstring_with_continuation.py"))]
+    #[test_case(Path::new("docstring_with_multiple_continuations.py"))]
     #[test_case(Path::new("docstring_with_semicolon.py"))]
     #[test_case(Path::new("empty.py"))]
     #[test_case(Path::new("existing_import.py"))]
@@ -832,6 +829,7 @@ mod tests {
     #[test_case(Path::new("docstring_followed_by_continuation.py"))]
     #[test_case(Path::new("docstring_only.py"))]
     #[test_case(Path::new("docstring_with_continuation.py"))]
+    #[test_case(Path::new("docstring_with_multiple_continuations.py"))]
     #[test_case(Path::new("docstring_with_semicolon.py"))]
     #[test_case(Path::new("empty.py"))]
     #[test_case(Path::new("existing_import.py"))]
