@@ -5,7 +5,7 @@ use ruff_python_ast as ast;
 
 use crate::Db;
 use crate::types::KnownClass;
-use crate::types::enums::enum_member_literals;
+use crate::types::enums::{enum_member_literals, enum_metadata};
 use crate::types::tuple::{Tuple, TupleLength, TupleType};
 
 use super::Type;
@@ -127,31 +127,39 @@ impl<'a, 'db> CallArguments<'a, 'db> {
     /// contains the same arguments, but with one or more of the argument types expanded.
     ///
     /// [argument type expansion]: https://typing.python.org/en/latest/spec/overload.html#argument-type-expansion
-    pub(crate) fn expand(
-        &self,
-        db: &'db dyn Db,
-    ) -> impl Iterator<Item = Vec<CallArguments<'a, 'db>>> + '_ {
+    pub(super) fn expand(&self, db: &'db dyn Db) -> impl Iterator<Item = Expansion<'a, 'db>> + '_ {
+        /// Maximum number of argument lists that can be generated in a single expansion step.
+        static MAX_EXPANSIONS: usize = 512;
+
         /// Represents the state of the expansion process.
+        enum State<'a, 'b, 'db> {
+            LimitReached(usize),
+            Expanding(ExpandingState<'a, 'b, 'db>),
+        }
+
+        /// Represents the expanding state with either the initial types or the expanded types.
         ///
         /// This is useful to avoid cloning the initial types vector if none of the types can be
         /// expanded.
-        enum State<'a, 'b, 'db> {
+        enum ExpandingState<'a, 'b, 'db> {
             Initial(&'b Vec<Option<Type<'db>>>),
             Expanded(Vec<CallArguments<'a, 'db>>),
         }
 
-        impl<'db> State<'_, '_, 'db> {
+        impl<'db> ExpandingState<'_, '_, 'db> {
             fn len(&self) -> usize {
                 match self {
-                    State::Initial(_) => 1,
-                    State::Expanded(expanded) => expanded.len(),
+                    ExpandingState::Initial(_) => 1,
+                    ExpandingState::Expanded(expanded) => expanded.len(),
                 }
             }
 
             fn iter(&self) -> impl Iterator<Item = &[Option<Type<'db>>]> + '_ {
                 match self {
-                    State::Initial(types) => Either::Left(std::iter::once(types.as_slice())),
-                    State::Expanded(expanded) => {
+                    ExpandingState::Initial(types) => {
+                        Either::Left(std::iter::once(types.as_slice()))
+                    }
+                    ExpandingState::Expanded(expanded) => {
                         Either::Right(expanded.iter().map(CallArguments::types))
                     }
                 }
@@ -160,44 +168,82 @@ impl<'a, 'db> CallArguments<'a, 'db> {
 
         let mut index = 0;
 
-        std::iter::successors(Some(State::Initial(&self.types)), move |previous| {
-            // Find the next type that can be expanded.
-            let expanded_types = loop {
-                let arg_type = self.types.get(index)?;
-                if let Some(arg_type) = arg_type {
-                    if let Some(expanded_types) = expand_type(db, *arg_type) {
-                        break expanded_types;
+        std::iter::successors(
+            Some(State::Expanding(ExpandingState::Initial(&self.types))),
+            move |previous| {
+                let state = match previous {
+                    State::LimitReached(index) => return Some(State::LimitReached(*index)),
+                    State::Expanding(expanding_state) => expanding_state,
+                };
+
+                // Find the next type that can be expanded.
+                let expanded_types = loop {
+                    let arg_type = self.types.get(index)?;
+                    if let Some(arg_type) = arg_type {
+                        if let Some(expanded_types) = expand_type(db, *arg_type) {
+                            break expanded_types;
+                        }
+                    }
+                    index += 1;
+                };
+
+                let expansion_size = expanded_types.len() * state.len();
+                if expansion_size > MAX_EXPANSIONS {
+                    tracing::debug!(
+                        "Skipping argument type expansion as it would exceed the \
+                    maximum number of expansions ({MAX_EXPANSIONS})"
+                    );
+                    return Some(State::LimitReached(index));
+                }
+
+                let mut expanded_arguments = Vec::with_capacity(expansion_size);
+
+                for pre_expanded_types in state.iter() {
+                    for subtype in &expanded_types {
+                        let mut new_expanded_types = pre_expanded_types.to_vec();
+                        new_expanded_types[index] = Some(*subtype);
+                        expanded_arguments.push(CallArguments::new(
+                            self.arguments.clone(),
+                            new_expanded_types,
+                        ));
                     }
                 }
+
+                // Increment the index to move to the next argument type for the next iteration.
                 index += 1;
-            };
 
-            let mut expanded_arguments = Vec::with_capacity(expanded_types.len() * previous.len());
-
-            for pre_expanded_types in previous.iter() {
-                for subtype in &expanded_types {
-                    let mut new_expanded_types = pre_expanded_types.to_vec();
-                    new_expanded_types[index] = Some(*subtype);
-                    expanded_arguments.push(CallArguments::new(
-                        self.arguments.clone(),
-                        new_expanded_types,
-                    ));
-                }
-            }
-
-            // Increment the index to move to the next argument type for the next iteration.
-            index += 1;
-
-            Some(State::Expanded(expanded_arguments))
-        })
+                Some(State::Expanding(ExpandingState::Expanded(
+                    expanded_arguments,
+                )))
+            },
+        )
         .skip(1) // Skip the initial state, which has no expanded types.
         .map(|state| match state {
-            State::Initial(_) => unreachable!("initial state should be skipped"),
-            State::Expanded(expanded) => expanded,
+            State::LimitReached(index) => Expansion::LimitReached(index),
+            State::Expanding(ExpandingState::Initial(_)) => {
+                unreachable!("initial state should be skipped")
+            }
+            State::Expanding(ExpandingState::Expanded(expanded)) => Expansion::Expanded(expanded),
         })
     }
 }
 
+/// Represents a single element of the expansion process for argument types for [`expand`].
+///
+/// [`expand`]: CallArguments::expand
+pub(super) enum Expansion<'a, 'db> {
+    /// Indicates that the expansion process has reached the maximum number of argument lists
+    /// that can be generated in a single step.
+    ///
+    /// The contained `usize` is the index of the argument type which would have been expanded
+    /// next, if not for the limit.
+    LimitReached(usize),
+
+    /// Contains the expanded argument lists, where each list contains the same arguments, but with
+    /// one or more of the argument types expanded.
+    Expanded(Vec<CallArguments<'a, 'db>>),
+}
+
 impl<'a, 'db> FromIterator<(Argument<'a>, Option<Type<'db>>)> for CallArguments<'a, 'db> {
     fn from_iter<T>(iter: T) -> Self
     where
@@ -208,10 +254,32 @@ impl<'a, 'db> FromIterator<(Argument<'a>, Option<Type<'db>>)> for CallArguments<
     }
 }
 
+/// Returns `true` if the type can be expanded into its subtypes.
+///
+/// In other words, it returns `true` if [`expand_type`] returns [`Some`] for the given type.
+pub(crate) fn is_expandable_type<'db>(db: &'db dyn Db, ty: Type<'db>) -> bool {
+    match ty {
+        Type::NominalInstance(instance) => {
+            let class = instance.class(db);
+            class.is_known(db, KnownClass::Bool)
+                || instance.tuple_spec(db).is_some_and(|spec| match &*spec {
+                    Tuple::Fixed(fixed_length_tuple) => fixed_length_tuple
+                        .all_elements()
+                        .any(|element| is_expandable_type(db, *element)),
+                    Tuple::Variable(_) => false,
+                })
+                || enum_metadata(db, class.class_literal(db).0).is_some()
+        }
+        Type::Union(_) => true,
+        _ => false,
+    }
+}
+
 /// Expands a type into its possible subtypes, if applicable.
 ///
 /// Returns [`None`] if the type cannot be expanded.
 fn expand_type<'db>(db: &'db dyn Db, ty: Type<'db>) -> Option<Vec<Type<'db>>> {
+    // NOTE: Update `is_expandable_type` if this logic changes accordingly.
     match ty {
         Type::NominalInstance(instance) => {
             let class = instance.class(db);
@@ -257,7 +325,7 @@ fn expand_type<'db>(db: &'db dyn Db, ty: Type<'db>) -> Option<Vec<Type<'db>>> {
 
             None
         }
-        Type::Union(union) => Some(union.iter(db).copied().collect()),
+        Type::Union(union) => Some(union.elements(db).to_vec()),
         // We don't handle `type[A | B]` here because it's already stored in the expanded form
         // i.e., `type[A] | type[B]` which is handled by the `Type::Union` case.
         _ => None,
