@@ -48,6 +48,13 @@ fn enclosing_generic_contexts<'db>(
                     .last_definition_signature(db)
                     .generic_context
             }
+            NodeWithScopeKind::TypeAlias(type_alias) => {
+                let definition = index.expect_single_definition(type_alias.node(module));
+                binding_type(db, definition)
+                    .into_type_alias()?
+                    .into_pep_695_type_alias()?
+                    .generic_context(db)
+            }
             _ => None,
         })
 }
@@ -134,6 +141,18 @@ impl<'db> GenericContext<'db> {
         Self::new(db, type_params.into_iter().collect::<FxOrderSet<_>>())
     }
 
+    /// Merge this generic context with another, returning a new generic context that
+    /// contains type variables from both contexts.
+    pub(crate) fn merge(self, db: &'db dyn Db, other: Self) -> Self {
+        Self::from_typevar_instances(
+            db,
+            self.variables(db)
+                .iter()
+                .chain(other.variables(db).iter())
+                .copied(),
+        )
+    }
+
     fn variable_from_type_param(
         db: &'db dyn Db,
         index: &'db SemanticIndex<'db>,
@@ -451,45 +470,50 @@ fn is_subtype_in_invariant_position<'db, C: Constraints<'db>>(
     derived_materialization: MaterializationKind,
     base_type: &Type<'db>,
     base_materialization: MaterializationKind,
+    visitor: &HasRelationToVisitor<'db, C>,
 ) -> C {
     let derived_top = derived_type.top_materialization(db);
     let derived_bottom = derived_type.bottom_materialization(db);
     let base_top = base_type.top_materialization(db);
     let base_bottom = base_type.bottom_materialization(db);
+
+    let is_subtype_of = |derived: Type<'db>, base: Type<'db>| {
+        derived.has_relation_to_impl(db, base, TypeRelation::Subtyping, visitor)
+    };
     match (derived_materialization, base_materialization) {
         // `Derived` is a subtype of `Base` if the range of materializations covered by `Derived`
         // is a subset of the range covered by `Base`.
-        (MaterializationKind::Top, MaterializationKind::Top) => C::from_bool(
-            db,
-            base_bottom.is_subtype_of(db, derived_bottom)
-                && derived_top.is_subtype_of(db, base_top),
-        ),
+        (MaterializationKind::Top, MaterializationKind::Top) => {
+            is_subtype_of(base_bottom, derived_bottom)
+                .and(db, || is_subtype_of(derived_top, base_top))
+        }
         // One bottom is a subtype of another if it covers a strictly larger set of materializations.
-        (MaterializationKind::Bottom, MaterializationKind::Bottom) => C::from_bool(
-            db,
-            derived_bottom.is_subtype_of(db, base_bottom)
-                && base_top.is_subtype_of(db, derived_top),
-        ),
+        (MaterializationKind::Bottom, MaterializationKind::Bottom) => {
+            is_subtype_of(derived_bottom, base_bottom)
+                .and(db, || is_subtype_of(base_top, derived_top))
+        }
         // The bottom materialization of `Derived` is a subtype of the top materialization
         // of `Base` if there is some type that is both within the
         // range of types covered by derived and within the range covered by base, because if such a type
         // exists, it's a subtype of `Top[base]` and a supertype of `Bottom[derived]`.
-        (MaterializationKind::Bottom, MaterializationKind::Top) => C::from_bool(
-            db,
-            (base_bottom.is_subtype_of(db, derived_bottom)
-                && derived_bottom.is_subtype_of(db, base_top))
-                || (base_bottom.is_subtype_of(db, derived_top)
-                    && derived_top.is_subtype_of(db, base_top)
-                    || (base_top.is_subtype_of(db, derived_top)
-                        && derived_bottom.is_subtype_of(db, base_top))),
-        ),
+        (MaterializationKind::Bottom, MaterializationKind::Top) => {
+            (is_subtype_of(base_bottom, derived_bottom)
+                .and(db, || is_subtype_of(derived_bottom, base_top)))
+            .or(db, || {
+                is_subtype_of(base_bottom, derived_top)
+                    .and(db, || is_subtype_of(derived_top, base_top))
+            })
+            .or(db, || {
+                is_subtype_of(base_top, derived_top)
+                    .and(db, || is_subtype_of(derived_bottom, base_top))
+            })
+        }
         // A top materialization is a subtype of a bottom materialization only if both original
         // un-materialized types are the same fully static type.
-        (MaterializationKind::Top, MaterializationKind::Bottom) => C::from_bool(
-            db,
-            derived_top.is_subtype_of(db, base_bottom)
-                && base_top.is_subtype_of(db, derived_bottom),
-        ),
+        (MaterializationKind::Top, MaterializationKind::Bottom) => {
+            is_subtype_of(derived_top, base_bottom)
+                .and(db, || is_subtype_of(base_top, derived_bottom))
+        }
     }
 }
 
@@ -503,23 +527,32 @@ fn has_relation_in_invariant_position<'db, C: Constraints<'db>>(
     base_type: &Type<'db>,
     base_materialization: Option<MaterializationKind>,
     relation: TypeRelation,
+    visitor: &HasRelationToVisitor<'db, C>,
 ) -> C {
     match (derived_materialization, base_materialization, relation) {
         // Top and bottom materializations are fully static types, so subtyping
         // is the same as assignability.
-        (Some(derived_mat), Some(base_mat), _) => {
-            is_subtype_in_invariant_position(db, derived_type, derived_mat, base_type, base_mat)
-        }
+        (Some(derived_mat), Some(base_mat), _) => is_subtype_in_invariant_position(
+            db,
+            derived_type,
+            derived_mat,
+            base_type,
+            base_mat,
+            visitor,
+        ),
         // Subtyping between invariant type parameters without a top/bottom materialization involved
         // is equivalence
-        (None, None, TypeRelation::Subtyping) => {
-            C::from_bool(db, derived_type.is_equivalent_to(db, *base_type))
-        }
-        (None, None, TypeRelation::Assignability) => C::from_bool(
-            db,
-            derived_type.is_assignable_to(db, *base_type)
-                && base_type.is_assignable_to(db, *derived_type),
-        ),
+        (None, None, TypeRelation::Subtyping) => derived_type.when_equivalent_to(db, *base_type),
+        (None, None, TypeRelation::Assignability) => derived_type
+            .has_relation_to_impl(db, *base_type, TypeRelation::Assignability, visitor)
+            .and(db, || {
+                base_type.has_relation_to_impl(
+                    db,
+                    *derived_type,
+                    TypeRelation::Assignability,
+                    visitor,
+                )
+            }),
         // For gradual types, A <: B (subtyping) is defined as Top[A] <: Bottom[B]
         (None, Some(base_mat), TypeRelation::Subtyping) => is_subtype_in_invariant_position(
             db,
@@ -527,6 +560,7 @@ fn has_relation_in_invariant_position<'db, C: Constraints<'db>>(
             MaterializationKind::Top,
             base_type,
             base_mat,
+            visitor,
         ),
         (Some(derived_mat), None, TypeRelation::Subtyping) => is_subtype_in_invariant_position(
             db,
@@ -534,6 +568,7 @@ fn has_relation_in_invariant_position<'db, C: Constraints<'db>>(
             derived_mat,
             base_type,
             MaterializationKind::Bottom,
+            visitor,
         ),
         // And A <~ B (assignability) is Bottom[A] <: Top[B]
         (None, Some(base_mat), TypeRelation::Assignability) => is_subtype_in_invariant_position(
@@ -542,6 +577,7 @@ fn has_relation_in_invariant_position<'db, C: Constraints<'db>>(
             MaterializationKind::Bottom,
             base_type,
             base_mat,
+            visitor,
         ),
         (Some(derived_mat), None, TypeRelation::Assignability) => is_subtype_in_invariant_position(
             db,
@@ -549,6 +585,7 @@ fn has_relation_in_invariant_position<'db, C: Constraints<'db>>(
             derived_mat,
             base_type,
             MaterializationKind::Top,
+            visitor,
         ),
     }
 }
@@ -590,7 +627,11 @@ impl<'db> Specialization<'db> {
         let new_specialization = self.apply_type_mapping(db, &TypeMapping::Specialization(other));
         match other.materialization_kind(db) {
             None => new_specialization,
-            Some(materialization_kind) => new_specialization.materialize(db, materialization_kind),
+            Some(materialization_kind) => new_specialization.materialize_impl(
+                db,
+                materialization_kind,
+                &ApplyTypeMappingVisitor::default(),
+            ),
         }
     }
 
@@ -608,6 +649,9 @@ impl<'db> Specialization<'db> {
         type_mapping: &TypeMapping<'a, 'db>,
         visitor: &ApplyTypeMappingVisitor<'db>,
     ) -> Self {
+        if let TypeMapping::Materialize(materialization_kind) = type_mapping {
+            return self.materialize_impl(db, *materialization_kind, visitor);
+        }
         let types: Box<[_]> = self
             .types(db)
             .iter()
@@ -684,10 +728,11 @@ impl<'db> Specialization<'db> {
         )
     }
 
-    pub(super) fn materialize(
+    pub(super) fn materialize_impl(
         self,
         db: &'db dyn Db,
         materialization_kind: MaterializationKind,
+        visitor: &ApplyTypeMappingVisitor<'db>,
     ) -> Self {
         // The top and bottom materializations are fully static types already, so materializing them
         // further does nothing.
@@ -705,14 +750,17 @@ impl<'db> Specialization<'db> {
                     TypeVarVariance::Bivariant => {
                         // With bivariance, all specializations are subtypes of each other,
                         // so any materialization is acceptable.
-                        vartype.materialize(db, MaterializationKind::Top)
+                        vartype.materialize(db, MaterializationKind::Top, visitor)
+                    }
+                    TypeVarVariance::Covariant => {
+                        vartype.materialize(db, materialization_kind, visitor)
                     }
-                    TypeVarVariance::Covariant => vartype.materialize(db, materialization_kind),
                     TypeVarVariance::Contravariant => {
-                        vartype.materialize(db, materialization_kind.flip())
+                        vartype.materialize(db, materialization_kind.flip(), visitor)
                     }
                     TypeVarVariance::Invariant => {
-                        let top_materialization = vartype.materialize(db, MaterializationKind::Top);
+                        let top_materialization =
+                            vartype.materialize(db, MaterializationKind::Top, visitor);
                         if !vartype.is_equivalent_to(db, top_materialization) {
                             has_dynamic_invariant_typevar = true;
                         }
@@ -723,7 +771,11 @@ impl<'db> Specialization<'db> {
             .collect();
         let tuple_inner = self.tuple_inner(db).and_then(|tuple| {
             // Tuples are immutable, so tuple element types are always in covariant position.
-            tuple.materialize(db, materialization_kind)
+            tuple.apply_type_mapping_impl(
+                db,
+                &TypeMapping::Materialize(materialization_kind),
+                visitor,
+            )
         });
         let new_materialization_kind = if has_dynamic_invariant_typevar {
             Some(materialization_kind)
@@ -790,6 +842,7 @@ impl<'db> Specialization<'db> {
                     other_type,
                     other_materialization_kind,
                     relation,
+                    visitor,
                 ),
                 TypeVarVariance::Covariant => {
                     self_type.has_relation_to_impl(db, *other_type, relation, visitor)
